import numpy as np
import pandas as pd
from tqdm import tqdm
from typing import Tuple

from pylenm2.data import fetchers
from pylenm2.stats import metrics
from pylenm2.stats import preprocess
from pylenm2.utils import constants as c

import logging
from pylenm2 import logger_config

transformation_logger = logger_config.setup_logging(
    module_name=__name__,
    # level=logging.INFO,
    level=logging.DEBUG,
    logfile_dir=c.LOGFILE_DIR,
)


def interpolate_station_data(
        # self, 
        data_pylenm_dm, 
        station_name, 
        analytes, 
        frequency='2W',
    ) -> pd.DataFrame:
    """Resamples the data based on the frequency specified and interpolates the values of the analytes.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        station_name (str): name of the station to be processed.
        analytes (list): list of analyte names to use
        frequency (str, optional): {‘D’, ‘W’, ‘M’, ‘Y’} frequency to interpolate. See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html for valid frequency inputs. (e.g. ‘W’ = every week, ‘D ’= every day, ‘2W’ = every 2 weeks). Defaults to '2W'.

    Returns:
        pd.DataFrame
    """
    # data = self.data
    data = data_pylenm_dm.data
    inter_series = {}
    query = data[data.STATION_ID == station_name]
    
    for analyte in analytes:
        series = query[query.ANALYTE_NAME == analyte]
        series = (series[['COLLECTION_DATE', 'RESULT']])
        series.COLLECTION_DATE = pd.to_datetime(series.COLLECTION_DATE)
        series.index = series.COLLECTION_DATE
        original_dates = series.index
        series = series.drop('COLLECTION_DATE', axis=1)
        series = series.rename({'RESULT': analyte}, axis=1)
        upsampled = series.resample(frequency).mean()
        interpolated = upsampled.interpolate(method='linear', order=2)
        inter_series[analyte] = interpolated
    
    join = inter_series[analytes[0]]
    join = join.drop(analytes[0], axis=1)
    
    for analyte in analytes:
        join = join.join(inter_series[analyte])
    
    join = join.dropna()
    
    return join


def interpolate_stations_by_analyte(
        data_pylenm_dm, 
        analyte, 
        frequency='2W', 
        rm_outliers=True, 
        z_threshold=3,
    ) -> pd.DataFrame:
    """Resamples analyte data based on the frequency specified and interpolates the values in between. NaN values are replaced with the average value per station.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        analyte (_type_): analyte name for interpolation of all present stations.
        frequency (str, optional): {‘D’, ‘W’, ‘M’, ‘Y’} frequency to interpolate. See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html for valid frequency inputs. (e.g. ‘W’ = every week, ‘D ’= every day, ‘2W’ = every 2 weeks). Defaults to '2W'.
        rm_outliers (bool, optional): flag to remove outliers in the data. Defaults to True.
        z_threshold (int, optional): z_score threshold to eliminate outliers. Defaults to 3.

    Returns:
        pd.DataFrame: interpolated data
    """
    # data = data_pylenm_dm.data

    df_t, dates = _transform_time_series( 
        data_pylenm_dm=data_pylenm_dm,
        analytes=[analyte], 
        resample=frequency, 
        rm_outliers=True, 
        z_threshold=z_threshold,
    )

    res_interp = fetchers._get_individual_analyte_df(
        data=df_t, 
        dates=dates, 
        analyte=analyte,
    )
    res_interp = res_interp.dropna(axis=1, how='all')
    
    return res_interp


# IN THE WORKS
def _transform_time_series(
        data_pylenm_dm, 
        analytes=[], 
        resample='2W', 
        rm_outliers=False, 
        z_threshold=4,
    ) -> Tuple[pd.DataFrame, pd.DatetimeIndex]:
    """<Function docstring> TODO: write function docstring.
    TODO: The function can be optimized a lot for a faster performance. Come back to this once everything is done.
    TODO: Get the logic confirmed by Zexuan.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        ... TODO: Write the args docstring.

    Returns:
        Tuple[pd.DataFrame, pd.DatetimeIndex]: Returns a tuple of the dataframe and the dates. TODO: Write more details about the returned data.
    """

    data = data_pylenm_dm.data
    
    def transform_time_series_by_analyte(data, analyte_name):
        """<Nested-Function docstring> TODO: write function docstring.

        Args:
            data (pd.DataFrame): input dataframe.
            analyte_name (str): name of the analyte.
        
        Returns:
            pd.DataFrame: return sample dataframe for the analyte.
        """
        # stations_analyte = np.unique(data[data.ANALYTE_NAME == analyte_name].STATION_ID)

        # # Create array of equally spaced dates
        # start_date = pd.Timestamp(data.COLLECTION_DATE.min())
        # end_date = pd.Timestamp(data.COLLECTION_DATE.max())
        # date_delta = (end_date - start_date) + pd.Timedelta(days=1)    # to include the end date as station
        # t = np.linspace(start_date.value, end_date.value, date_delta.days)
        # t = pd.to_datetime(t).date
        # # t = pd.Series(t)
        # # t = t.apply(lambda x: x.replace(minute=0, hour=0, second=0, microsecond=0, nanosecond=0))

        # condensed = data[data.ANALYTE_NAME == analyte_name].groupby(['STATION_ID','COLLECTION_DATE']).mean()    # NOTE: Breaks the code
        # condensed = data[data.ANALYTE_NAME == analyte_name].groupby(['STATION_ID','COLLECTION_DATE'])['RESULT'].mean().to_frame('RESULT')     # NOTE: Works. Result must have (station, date) as index
        condensed = data[data.ANALYTE_NAME == analyte_name].groupby(['STATION_ID','COLLECTION_DATE'], as_index=False)['RESULT'].mean()  # NOTE: Much better approach.

        analyte_df_resample = condensed.pivot(columns="COLLECTION_DATE", index="STATION_ID", values="RESULT")

        # analyte_df_resample = pd.DataFrame(index=stations_analyte, columns=t)
        analyte_df_resample.sort_index(inplace=True)
        
        # for station in stations_analyte:    # NOTE: Performs pivot. Implemented more efficiently above.
        #     for date in condensed.loc[station].index:
        #         analyte_df_resample.at[station, pd.to_datetime(date).date()] = condensed.loc[station,date].RESULT
        
        analyte_df_resample = analyte_df_resample.astype('float').T
        analyte_df_resample = analyte_df_resample.interpolate(method='linear')
        return analyte_df_resample


    # data_analyte_groups = data.groupby("ANALYTE_NAME", as_index=False)
    # transformed_ts_data_analyte = data_analyte_groups.apply(
    #     lambda subdf: transform_time_series_by_analyte(
    #         data=subdf, analyte_name=subdf.ANALYTE_NAME.iloc[0]
    #     )
    # )


    # Save each analyte data
    cutoff_dates = []
    analyte_data = []
    for analyte in analytes:
        ana_data = transform_time_series_by_analyte(data, analyte)

        if(rm_outliers):
            col_num = ana_data.shape[1]
            for col in range(col_num):
                try:
                    ana_data.iloc[:,col] = preprocess.remove_outliers(
                        ana_data.iloc[:,col], 
                        z_threshold=z_threshold,
                    )
                except Exception as e:
                    transformation_logger.error(e)

            ana_data = ana_data.interpolate(method='linear')
        
        ana_data.index = pd.to_datetime(ana_data.index)
        
        # Resample
        ana_data_resample = ana_data.resample(resample).mean()
        
        # Save data
        analyte_data.append(ana_data_resample)
        
        # Determine cuttoff point for number of NaNs in dataset
        passes_limit = []
        for date in ana_data_resample.index:
            limit = 0.7 * ana_data_resample.shape[1]
            curr = ana_data_resample.isna().loc[date,:].value_counts()
            if('False' in str(curr)):
                curr_total = ana_data_resample.isna().loc[date,:].value_counts()[0]
                if curr_total > limit:
                    passes_limit.append(date)
        passes_limit = pd.to_datetime(passes_limit)
        
        cutoff_dates.append(passes_limit.min())
    
    start_index = pd.Series(cutoff_dates).max()

    # Get list of shared stations amongst all the listed analytes
    combined_station_list = []
    for x in range(len(analytes)):
        combined_station_list = combined_station_list + list(analyte_data[x].columns)
    
    combined_count = pd.Series(combined_station_list).value_counts()
    shared_stations = list(
        combined_count[
            list(pd.Series(combined_station_list).value_counts()==len(analytes))
        ].index
    )

    # Vectorize data
    vectorized_df = pd.DataFrame(columns=analytes, index=shared_stations)

    for analyte, num in zip(analytes, range(len(analytes))):
        for station in shared_stations:
            analyte_data_full = analyte_data[num][station].fillna(analyte_data[num][station].mean())
            vectorized_df.at[station, analyte] = analyte_data_full[start_index:].values

    dates = ana_data_resample[start_index:].index
    
    return vectorized_df, dates


def add_dist_to_source(
        XX, 
        source_coordinate=c.DEFAULT_SOURCE_COORDINATES, 
        col_name='dist_to_source',
    ) -> pd.DataFrame:
    """adds column to data with the distance of a record to the source coordinate

    Args:
        XX (pd.DataFrame): data with coordinate information
        source_coordinate (list, optional): source coordinate. Defaults to [436642.70,3681927.09].
        col_name (str, optional): name to assign new column. Defaults to 'dist_to_source'.

    Returns:
        pd.DataFrame: returns original data with additional column with the distance.
    """
    x1,y1 = source_coordinate
    distances = []
    for i in range(XX.shape[0]):
        x2,y2 = XX.iloc[i][0], XX.iloc[i][1]
        # distances.append(self.dist([x1,y1],[x2,y2]))
        distances.append(metrics.dist([x1,y1],[x2,y2]))
    XX[col_name] = distances
    return XX


## by K. Whiteaker, kwhit@alum.mit.edu
def time_average_all_stations(
        # self, 
        data_pylenm_dm, 
        analyte, 
        period='1W', 
        rm_outliers=True, 
        std_thresh=2.2, 
        lowess_frac=0.1,
    ):
    """Transforms all analyte data from exact measurements, at potentially 
    different dates for each station, into periodic averaged data (ex. weekly 
    averages) at the same dates for all stations.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        analyte (_type_): analyte name
        period (str, optional): {‘D’, ‘W’, ‘M’, ‘Y’} time period over which to average. See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html for valid frequency inputs. (e.g. ‘2W’ = every 2 weeks). Defaults to '1W'
        rm_outliers (bool, optional): flag to remove outliers in the data via LOWESS fit. Defaults to True
        std_thresh (int, optional): number of standard deviations in (observation - fit) outside of which is considered an outlier. Defaults to 2.2
        lowess_frac (float, optional): fraction of total data points considered in local fits of LOWESS smoother. A smaller value captures more local behaviour, and may be required for large datasets. Defaults to 0.1

    Returns:
        pd.DataFrame: dataframe with columns = stations, rows = dates, and data representing the average for that station between each date
    """
    # import data and filter for the chosen analyte
    data_concentration = data_pylenm_dm.data
    data_concentration = data_concentration[data_concentration.ANALYTE_NAME == analyte]
    data_construction = data_pylenm_dm.construction_data

    # create reshaped_df with columns=all available stations and rows=all measurement dates, to iterate over when conducting periodic averaging
    stations = np.unique(data_concentration[data_concentration.ANALYTE_NAME == analyte].STATION_ID)
    times = pd.to_datetime(np.unique(data_concentration.COLLECTION_DATE))
    reshaped_df = pd.DataFrame(index=times, columns=stations)
    
    # remove outliers from each station's data using remove_outlier_lowess(), and save the remaining data in reshaped_df
    for station in stations:
        data_in = data_concentration[data_concentration.STATION_ID == station].set_index('COLLECTION_DATE').RESULT
        data_in.index = pd.to_datetime(data_in.index)
        if rm_outliers:
            data_xOutliers = preprocess.remove_outliers_lowess(
                data_in, lowess_frac, std_thresh, 
            )
        else:
            data_xOutliers = data_in
        reshaped_df[station].loc[data_xOutliers.index] = data_xOutliers
    
    # find first and last measurement dates
    date_start = pd.Timestamp(times.min())
    date_end = pd.Timestamp(times.max())
    # make an array of datetimes, 1 period (ex. 1 week) apart, that spanning from before the first measurement date to after the last measurement date
    period_timesteps = pd.date_range(
        min(times) - pd.Timedelta(period), 
        max(times) + pd.Timedelta(period), 
        freq=period
    ).to_numpy()

    # conduct periodic averaging: take periodic (ex. 1 week) averages of each station and save the result to averaged_df
    averaged_df = pd.DataFrame(
        index=period_timesteps, 
        columns=stations, 
        dtype=float,
    )
    
    for period_step in tqdm(range(period_timesteps.size-1), desc="Timesteps", total=period_timesteps.size-1):  # for each period...
        # if period_step%100==0:
        #     print(period_step)  # print progress
        period_step_bin = []
        relevant_times = times[(times > period_timesteps[period_step]) & (times < period_timesteps[period_step+1])]
        
        for relevant_time in relevant_times:  # for each existing measurement that occurred within this period...
            timestep_data = reshaped_df.loc[relevant_time]  # identify all data points at this timestep
            period_step_bin.append(timestep_data)
        
        if len(period_step_bin)>0:  # if there are any measurements in this period
            period_data = pd.concat(period_step_bin, axis=1)
            period_station_avg = period_data.mean(axis=1, skipna=True)  # for each station, take the average over this period
            averaged_df.loc[period_timesteps[period_step]] = period_station_avg
    
    return averaged_df


import os
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

from pylenm2.data import filters, fetchers, transformation
from pylenm2.utils import constants as c

import logging
from pylenm2 import logger_config

plot_pca_logger = logger_config.setup_logging(
    module_name=__name__,
    # level=logging.INFO,
    level=logging.DEBUG,
    logfile_dir=c.LOGFILE_DIR,
)


def plot_PCA_by_date(
        # self, 
        data_pylenm_dm,
        date, 
        analytes, 
        lag=0, 
        n_clusters=4, 
        return_clusters=False, 
        min_samples=3, 
        show_labels=True, 
        save_dir='plot_PCA_by_date', 
        filter=False, 
        col=None, 
        equals=[],
    ):
    """Generates a PCA biplot (PCA score plot + loading plot) of the data given a date in the dataset. The data is also clustered into n_clusters.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        date (str): date to be analyzed
        analytes (str): list of analyte names to use
        lag (int, optional): number of days to look ahead and behind the specified date (+/-). Defaults to 0.
        n_clusters (int, optional): number of clusters to split the data into.. Defaults to 4.
        return_clusters (bool, optional): Flag to return the cluster data to be used for spatial plotting.. Defaults to False.
        min_samples (int, optional): minimum number of samples the result should contain in order to execute.. Defaults to 3.
        show_labels (bool, optional): choose whether or not to show the name of the stations.. Defaults to True.
        save_dir (str, optional): name of the directory you want to save the plot to. Defaults to 'plot_PCA_by_date'.
        filter (bool, optional): flag to indicate filtering. Defaults to False.
        col (str, optional): column to filter. Example: col='STATION_ID'. Defaults to None.
        equals (list, optional): values to filter col by. Examples: equals=['FAI001A', 'FAI001B']. Defaults to [].
    """
    
    if(lag==0):
        
        # Prepare data
        # data = self.data
        data = data_pylenm_dm.data
        # data = self.simplify_data(data=data)
        data = filters.simplify_data(data=data)
        query = data[data.COLLECTION_DATE == date]
        
        # Filter data (if specified)
        if(filter):
            # filter_res = self.filter_by_column(data=self.construction_data, col=col, equals=equals)
            filter_res = filters.filter_by_column(
                data=data_pylenm_dm.construction_data, 
                col=col, 
                equals=equals,
            )
            
            # if('ERROR:' in str(filter_res)):
            #     return filter_res
            if filter_res is None:
                return None

            query_stations = list(query.STATION_ID.unique())
            filter_stations = list(filter_res.index.unique())
            intersect_stations = list(set(query_stations) & set(filter_stations))
            if len(intersect_stations) <= 0:
                plot_pca_logger.error("ERROR: No results for this query with the specifed filter parameters.")
                # return 'ERROR: No results for this query with the specifed filter parameters.'
                return None
            
            query = query[query['STATION_ID'].isin(intersect_stations)]
        
        a = list(np.unique(query.ANALYTE_NAME.values))  # get all analytes from dataset
        
        for value in analytes:
            if((value in a)==False):
                plot_pca_logger.error(f"ERROR: No analyte named {value} in data.")
                # return 'ERROR: No analyte named "{}" in data.'.format(value)
                return None
        
        analytes = sorted(analytes)
        query = query.loc[query.ANALYTE_NAME.isin(analytes)]

        if query.shape[0] == 0:
            plot_pca_logger.error(f"ERROR: {date} has no data for the 6 analytes.")
            # return 'ERROR: {} has no data for the 6 analytes.'.format(date)
            return None
        
        samples = query[['COLLECTION_DATE', 'STATION_ID', 'ANALYTE_NAME']].duplicated().value_counts()[0]
        
        # Check if there are enough samples
        if samples < min_samples:
            plot_pca_logger.error(f"ERROR: {date} does not have at least {min_samples} samples.")
            # return 'ERROR: {} does not have at least {} samples.'.format(date, min_samples)
            return None

        # if(len(np.unique(query.ANALYTE_NAME.values)) < 6):
        #     return 'ERROR: {} has less than the 6 analytes we want to analyze.'.format(date)
        else:
            # analytes = preprocess._custom_analyte_sort(np.unique(query.ANALYTE_NAME.values))
            analytes = sorted(analytes)
            piv = query.reset_index().pivot_table(
                index='STATION_ID', 
                columns='ANALYTE_NAME', 
                values='RESULT', 
                aggfunc=np.mean,
            )
            
            # return piv
    else:
        # If the data has already been calculated with the lag specified, retrieve it
        # if(self.jointData_is_set(lag=lag)==True): 
        #     data = self.__jointData[0]
        if data_pylenm_dm.is_set_jointData(lag=lag):
            data = data_pylenm_dm.jointData[0]
        
        # Otherwise, calculate it
        else:
            # data = self.getJointData(analytes, lag=lag)
            data = fetchers.getJointData(
                data_pylenm_dm=data_pylenm_dm, 
                analytes=analytes, 
                lag=lag,
            )
            # self.__set_jointData(data=data, lag=lag)
            data_pylenm_dm.set_jointData(data=data, lag=lag)
        
        # get new range based on the lag and create the pivor table to be able to do the correlation
        # dateStart, dateEnd = self.__getLagDate(date, lagDays=lag)
        dateStart, dateEnd = fetchers._getLagDate(date, lagDays=lag)
        # dateRange_key = str(dateStart.date()) + " - " + str(dateEnd.date())
        dateRange_key = f"{str(dateStart.date())} - {str(dateEnd.date())}"
        piv = pd.DataFrame(data.loc[dateRange_key]).unstack().T
        piv.index = piv.index.droplevel()
        piv = pd.DataFrame(piv).dropna(axis=0, how='all')
        num_NaNs = int(piv.isnull().sum().sum())
        samples = (piv.shape[0]*piv.shape[1])-num_NaNs
        for col in piv.columns:
            piv[col] = piv[col].astype('float64', errors='raise')
        if lag > 0:
            date = dateRange_key
        # return piv

    main_data = piv.dropna()
    
    # Fit PCA and KMeans models
    scaler = StandardScaler()
    X = scaler.fit_transform(main_data)
    
    pca = PCA(n_components=2)
    x_new = pca.fit_transform(X)
    
    pca_points = pd.DataFrame(x_new, columns=["x1", "x2"])
    k_Means = KMeans(n_init='auto', n_clusters=n_clusters, random_state=42)
    model = k_Means.fit(pca_points[['x1', 'x2']])
    predict = model.predict(pca_points[['x1', 'x2']])
    
    # attach predicted cluster to original points
    pca_points['predicted'] = model.labels_
    
    # Create a dataframe for cluster_centers (centroids)
    centroids = pd.DataFrame(model.cluster_centers_, columns=["x1", "x2"])
    colors = [
        'red', 'blue', 'orange', 'purple', 'green', 'beige', 
        'pink', 'black', 'cadetblue', 'lightgreen',
    ]
    pca_points['color'] = pca_points['predicted'].map(lambda p: colors[p])

    fig, ax = plt.subplots(figsize=(10,10))
    # ax = plt.axes()   # NOTE: Causes redundant overlapping x and y ticks.

    small_fontSize = 15
    large_fontSize = 20
    plt.rc('axes', titlesize=large_fontSize)
    plt.rc('axes', labelsize=large_fontSize)
    plt.rc('legend', fontsize=small_fontSize)
    plt.rc('xtick', labelsize=small_fontSize)
    plt.rc('ytick', labelsize=small_fontSize)

    def myplot(score, coeff, labels=None, c='r', centroids=None):
        xs = score.iloc[:,0]
        ys = score.iloc[:,1]
        n = coeff.shape[0]
        scalex = 1.0/(xs.max() - xs.min())
        scaley = 1.0/(ys.max() - ys.min())
        scatt_X = xs * scalex
        scatt_Y = ys * scaley
        scatter = plt.scatter(
            scatt_X, 
            scatt_Y, 
            alpha=0.8, 
            label='Stations', 
            c=c,
        )
        centers = plt.scatter(
            centroids.iloc[:,0] * scalex, 
            centroids.iloc[:,1] * scaley, 
            c=colors[0:n_clusters], 
            marker='X', 
            s=550,
        )

        for i in range(n):
            arrow = plt.arrow(
                0, 0, coeff[i,0], coeff[i,1], color='r', alpha=0.9, 
                head_width=0.05, head_length=0.05, label='Loadings',
            )
            if labels is None:
                plt.text(
                    coeff[i,0]* 1.15, coeff[i,1] * 1.15, f"Var{i+1}", 
                    color='g', ha='center', va='center',
                )
            else:
                plt.text(
                    coeff[i,0]* 1.15, coeff[i,1] * 1.15, labels[i], 
                    color='g', ha='center', va='bottom',
                )

        if(show_labels):
            for x_pos, y_pos, label in zip(scatt_X, scatt_Y, main_data.index):
                ax.annotate(
                    label, # The label for this point
                    xy=(x_pos, y_pos), # Position of the corresponding point
                    xytext=(7, 0),     # Offset text by 7 points to the right
                    textcoords='offset points', # tell it to use offset points
                    ha='left',         # Horizontally aligned to the left
                    va='center',       # Vertical alignment is centered
                    color='black', 
                    alpha=0.8,
                )
        plt.legend(
            [scatter, centers, arrow], 
            ['Stations', 'Station centroids','Loadings'],
        )

    samples = x_new.shape[0] * piv.shape[1]

    # Add the meta-data information box
    props = dict(boxstyle='round', facecolor='grey', alpha=0.15)
    ax.text(
        1.1, 
        0.5, 
        # 'Date:  {}\n\nSamples:          {}\nStations:               {}'.format(date, samples, x_new.shape[0]), 
        f"Date: {date}\n\nSamples: {samples}\nStations: {x_new.shape[0]}",
        transform=ax.transAxes, 
        fontsize=20, 
        fontweight='bold', 
        verticalalignment='bottom', 
        bbox=props,
    )

    plt.xlim(-1,1)
    plt.ylim(-1,1)
    plt.xlabel("PC{}".format(1))
    plt.ylabel("PC{}".format(2))
    # ax.set_title('PCA Biplot - ' + date, fontweight='bold')
    ax.set_title(f"PCA Biplot - {date}", fontweight='bold')
    plt.grid(alpha=0.5)

    #Call the function. Use only the 2 Principal Components.
    myplot(
        pca_points, 
        np.transpose(pca.components_[0:2, :]), 
        labels=piv.columns, 
        c=pca_points['color'], 
        centroids=centroids,
    )
    plt.show()
    
    # Save figure (if specified)
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    fig.savefig(
        f"{save_dir}/PCA Biplot - {date}.png", bbox_inches="tight",
        # save_dir + '/' + 'PCA Biplot - '+ date +'.png', bbox_inches="tight"
    )
    
    # Return clusters (if specified)
    if(return_clusters):
        stations = list(main_data.index)
        color_stations = list(pca_points.color)
        
        def merge(list1, list2): 
            merged_list = [(list1[i], list2[i]) for i in range(0, len(list1))] 
            return merged_list
        
        color_df = pd.DataFrame(
            merge(stations, color_stations), 
            columns=['STATION_ID', 'color'],
        )
        
        # if(self.get_Construction_Data==None):
        if data_pylenm_dm.get_construction_data is None:
            plot_pca_logger.error("You need to set the GPS data first for using the get_construction_data() function.")
            # print('You need to set the GPS data first using the getConstructionData function.')
            return None
        
        else:
            gps_color = pd.merge(
                data_pylenm_dm.get_construction_data(), 
                color_df, 
                on=['STATION_ID'],
            )
            return gps_color


def plot_PCA_by_year(
        # self, 
        data_pylenm_dm,
        year, 
        analytes, 
        n_clusters=4, 
        return_clusters=False, 
        min_samples=10, 
        show_labels=True, 
        save_dir='plot_PCA_by_year', 
        filter=False, 
        col=None, 
        equals=[],
    ):
    """Gernates a PCA biplot (PCA score plot + loading plot) of the data given a year in the dataset. The data is also clustered into n_clusters.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        year (int): year to be analyzed
        analytes (str): list of analyte names to use
        n_clusters (int, optional): number of clusters to split the data into.. Defaults to 4.
        return_clusters (bool, optional): Flag to return the cluster data to be used for spatial plotting.. Defaults to False.
        min_samples (int, optional): minimum number of samples the result should contain in order to execute.. Defaults to 3.
        show_labels (bool, optional): choose whether or not to show the name of the stations.. Defaults to True.
        save_dir (str, optional): name of the directory you want to save the plot to. Defaults to 'plot_PCA_by_date'.
        filter (bool, optional): flag to indicate filtering. Defaults to False.
        col (str, optional): column to filter. Example: col='STATION_ID'. Defaults to None.
        equals (list, optional): values to filter col by. Examples: equals=['FAI001A', 'FAI001B']. Defaults to [].
    """

    # Prepare data
    # data = self.data
    data = data_pylenm_dm.data
    # query = self.simplify_data(data=data)
    query = filters.simplify_data(data=data)
    query.COLLECTION_DATE = pd.to_datetime(query.COLLECTION_DATE)
    query = query[query.COLLECTION_DATE.dt.year == year]
    
    # Filter data (if specified)
    if(filter):
        # filter_res = self.filter_by_column(data=self.construction_data, col=col, equals=equals)
        filter_res = filters.filter_by_column(
            data=data_pylenm_dm.construction_data, 
            col=col, 
            equals=equals,
        )
        
        # if('ERROR:' in str(filter_res)):
        #     return filter_res
        if filter_res is None:
            return None

        query_stations = list(query.STATION_ID.unique())
        filter_stations = list(filter_res.index.unique())
        intersect_stations = list(set(query_stations) & set(filter_stations))

        if len(intersect_stations) <= 0:
            plot_pca_logger.error("ERROR: No results for this query with the specifed filter parameters.")
            # return 'ERROR: No results for this query with the specifed filter parameters.'
            return None
        
        query = query[query['STATION_ID'].isin(intersect_stations)]
    
    a = list(np.unique(query.ANALYTE_NAME.values))  # get all analytes from dataset
    
    for value in analytes:
        if((value in a)==False):
            plot_pca_logger.error(f"ERROR: No analyte named {value} in data.")
            # return 'ERROR: No analyte named "{}" in data.'.format(value)
            return None
    
    analytes = sorted(analytes)
    query = query.loc[query.ANALYTE_NAME.isin(analytes)]

    if(query.shape[0] == 0):
        plot_pca_logger.error(f"ERROR: {year} has no data for the 6 analytes.")
        # return 'ERROR: {} has no data for the 6 analytes.'.format(year)
        return None
    
    samples = query[['COLLECTION_DATE', 'STATION_ID', 'ANALYTE_NAME']].duplicated().value_counts()[0]

    if samples < min_samples:
        plot_pca_logger.error(f"ERROR: {year} does not have at least {min_samples} samples.")
        # return 'ERROR: {} does not have at least {} samples.'.format(year, min_samples)
        return None

    # if(len(np.unique(query.ANALYTE_NAME.values)) < 6):
    #     return 'ERROR: {} has less than the 6 analytes we want to analyze.'.format(year)
    else:
        # analytes = preprocess._custom_analyte_sort(np.unique(query.ANALYTE_NAME.values))
        analytes = sorted(analytes)
        piv = query.reset_index().pivot_table(
            index='STATION_ID', 
            columns='ANALYTE_NAME', 
            values='RESULT', 
            aggfunc=np.mean,
        )
        
        main_data = piv.dropna()
        # # FILTERING CODE
        # if(filter):
        #     res_stations = self.filter_stations(filter_station_by)
        #     main_data = main_data.loc[main_data.index.isin(res_stations)]
        
        # Fit PCA and KMeans models
        scaler = StandardScaler()
        X = scaler.fit_transform(main_data)
        
        pca = PCA(n_components=2)
        x_new = pca.fit_transform(X)

        pca_points = pd.DataFrame(x_new, columns=["x1", "x2"])
        k_Means = KMeans(n_init='auto', n_clusters=n_clusters, random_state=42)
        model = k_Means.fit(pca_points[['x1', 'x2']])
        predict = model.predict(pca_points[['x1', 'x2']])
        
        # attach predicted cluster to original points
        pca_points['predicted'] = model.labels_
        
        # Create a dataframe for cluster_centers (centroids)
        centroids = pd.DataFrame(model.cluster_centers_, columns=["x1", "x2"])
        colors = [
            'red', 'blue', 'orange', 'purple', 'green', 'beige', 
            'pink', 'black', 'cadetblue', 'lightgreen',
        ]
        pca_points['color'] = pca_points['predicted'].map(lambda p: colors[p])
        
        fig, ax = plt.subplots(figsize=(15,15))
        # ax = plt.axes()   # NOTE: Causes redundant overlapping x and y ticks.

        small_fontSize = 15
        large_fontSize = 20
        plt.rc('axes', titlesize=large_fontSize)
        plt.rc('axes', labelsize=large_fontSize)
        plt.rc('legend', fontsize=small_fontSize)
        plt.rc('xtick', labelsize=small_fontSize)
        plt.rc('ytick', labelsize=small_fontSize) 

        def myplot(score,coeff,labels=None,c='r', centroids=None):
            xs = score[:,0]
            ys = score[:,1]
            n = coeff.shape[0]
            scalex = 1.0/(xs.max() - xs.min())
            scaley = 1.0/(ys.max() - ys.min())
            scatt_X = xs * scalex
            scatt_Y = ys * scaley
            scatter = plt.scatter(
                scatt_X, 
                scatt_Y, 
                alpha=0.8, 
                label='Stations', 
                c=c,
            )
            centers = plt.scatter(
                centroids.iloc[:,0] * scalex, 
                centroids.iloc[:,1] * scaley, 
                c = colors[0:n_clusters], 
                marker='X', 
                s=550,
            )

            for i in range(n):
                arrow = plt.arrow(
                    0, 0, coeff[i,0], coeff[i,1], color='r', alpha=0.9, 
                    head_width=0.05, head_length=0.05,
                )
                if labels is None:
                    plt.text(
                        coeff[i,0] * 1.15, coeff[i,1] * 1.15, f"Var{i+1}", 
                        color='g', ha='center', va='center',
                    )
                else:
                    plt.text(
                        coeff[i,0] * 1.15, coeff[i,1] * 1.15, labels[i], 
                        color='g', ha='center', va='bottom',
                    )

            if(show_labels):
                for x_pos, y_pos, label in zip(scatt_X, scatt_Y, main_data.index):
                    ax.annotate(
                        label, # The label for this point
                        xy=(x_pos, y_pos), # Position of the corresponding point
                        xytext=(7, 0),     # Offset text by 7 points to the right
                        textcoords='offset points', # tell it to use offset points
                        ha='left',         # Horizontally aligned to the left
                        va='center', 
                        color='black', 
                        alpha=0.8,
                    )       # Vertical alignment is centered
            plt.legend(
                [scatter, centers, arrow], 
                ['Stations', 'Station centroids','Loadings'],
            )

        samples = x_new.shape[0]*piv.shape[1]    
        
        # Add meta-data information box
        props = dict(boxstyle='round', facecolor='grey', alpha=0.15)
        ax.text(
            1.1, 
            0.5, 
            # 'Date:  {}\n\nSamples:          {}\nStations:               {}'.format(year,samples, x_new.shape[0]), 
            f"Year: {year}\n\nSamples: {samples}\nStations: {x_new.shape[0]}",
            transform=ax.transAxes, 
            fontsize=20, 
            fontweight='bold', 
            verticalalignment='bottom', 
            bbox=props,
        )

        plt.xlim(-1,1)
        plt.ylim(-1,1)
        plt.xlabel("PC{}".format(1))
        plt.ylabel("PC{}".format(2))
        # ax.set_title('PCA Biplot - ' + str(year), fontweight='bold')
        ax.set_title(f"PCA Biplot - {str(year)}", fontweight='bold')
        plt.grid(alpha=0.5)

        #Call the function. Use only the 2 PCs.
        myplot(
            x_new[:,0:2], 
            np.transpose(pca.components_[0:2, :]), 
            labels=piv.columns, 
            c=pca_points['color'], 
            centroids=centroids,
        )
        plt.show()

        # Save figure (if specified)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        fig.savefig(
            f"{save_dir}/PCA Biplot - {year}.png", bbox_inches="tight",
            # save_dir + '/' + 'PCA Biplot - '+ str(year) +'.png', bbox_inches="tight",
        )
        
        # Return clusters (if specified)
        if(return_clusters):
            stations = list(main_data.index)
            color_stations = list(pca_points.color)
            
            def merge(list1, list2): 
                merged_list = [
                    (list1[i], list2[i]) 
                    for i in range(0, len(list1))
                ] 
                return merged_list
            
            color_df = pd.DataFrame(
                merge(stations, color_stations), 
                columns=['STATION_ID', 'color'],
            )
            
            # if(self.get_Construction_Data==None):
            if data_pylenm_dm.get_construction_data is None:
                plot_pca_logger.error('You need to set the GPS data first for using the setConstructionData function.')
                # print('You need to set the GPS data first for using the setConstructionData function.')
                return None
            
            else:
                gps_color = pd.merge(
                    data_pylenm_dm.get_construction_data(), 
                    color_df, 
                    on=['STATION_ID'],
                )
                return gps_color


def plot_PCA_by_station(
        # self, 
        data_pylenm_dm, 
        station_name, 
        analytes, 
        interpolate=False, 
        frequency='2W', 
        min_samples=10, 
        show_labels=True, 
        save_dir='plot_PCA_by_station',
    ):
    """Gernates a PCA biplot (PCA score plot + loading plot) of the data given a station_name in the dataset. Only uses the 6 important analytes.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        station_name (str): name of the station to be processed
        analytes (str): list of analyte names to use
        interpolate (bool, optional): choose to interpolate the data. Defaults to False.
        frequency (str, optional): {‘D’, ‘W’, ‘M’, ‘Y’} frequency to interpolate. See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html for valid frequency inputs. (e.g. ‘W’ = every week, ‘D ’= every day, ‘2W’ = every 2 weeks). Defaults to '2W'.
        min_samples (int, optional): minimum number of samples the result should contain in order to execute.. Defaults to 3.
        show_labels (bool, optional): choose whether or not to show the name of the stations.. Defaults to True.
        save_dir (str, optional): name of the directory you want to save the plot to. Defaults to 'plot_PCA_by_date'.
    """

    # data = self.data
    data = data_pylenm_dm.data
    query = data[data.STATION_ID == station_name]
    a = list(np.unique(query.ANALYTE_NAME.values))# get all analytes from dataset
    
    for value in analytes:
        if((value in a)==False):
            plot_pca_logger.error(f"ERROR: No analyte named {value} in data.")
            # return 'ERROR: No analyte named "{}" in data.'.format(value)
            return None
    
    analytes = sorted(analytes)

    query = query.loc[query.ANALYTE_NAME.isin(analytes)]
    x = query[['COLLECTION_DATE', 'ANALYTE_NAME']]
    unique = ~x.duplicated()
    query = query[unique]
    piv = query.reset_index().pivot(index='COLLECTION_DATE',columns='ANALYTE_NAME', values='RESULT')
    piv = piv[analytes]
    piv.index = pd.to_datetime(piv.index)
    totalSamples = piv.stack().shape[0]
    piv = piv.dropna()
    
    # Interpolate (if specified)
    if(interpolate):
        # piv = self.interpolate_station_data(station_name, analytes, frequency=frequency)
        piv = transformation.interpolate_station_data(
            station_name, 
            analytes, 
            frequency=frequency,
        )
        # title = 'PCA Biplot - ' + station_name + ' - interpolated every ' + frequency
        title = f"PCA Biplot - {station_name} - interpolated every {frequency}."
    else:
        # title = 'PCA Biplot - ' + station_name
        title = f"PCA Biplot - {station_name}"

    if query.shape[0] == 0:
        plot_pca_logger.error(f"ERROR: {date} has no data for the 6 analytes.")
        # return 'ERROR: {} has no data for the 6 analytes.'.format(date)
        return None
    
    samples = query[['COLLECTION_DATE', 'STATION_ID', 'ANALYTE_NAME']].duplicated().value_counts()[0]
    if samples < min_samples:
        plot_pca_logger.error(f"ERROR: {date} does not have at least {min_samples} samples.")
        # return 'ERROR: {} does not have at least {} samples.'.format(date, min_samples)
        return None
    
    # if(len(np.unique(query.ANALYTE_NAME.values)) < 6):
    #     return 'ERROR: {} has less than the 6 analytes we want to analyze.'.format(station_name)
    else:
        
        # Scale the values
        scaler = StandardScaler()
        X = scaler.fit_transform(piv.dropna())
        
        # Fit PCA model
        pca = PCA(n_components=2)
        x_new = pca.fit_transform(X)

        # Generate Plots
        fig, ax = plt.subplots(figsize=(15,15))
        # ax = plt.axes()   # NOTE: Causes redundance overlapping x and y ticks.

        small_fontSize = 15
        large_fontSize = 20
        plt.rc('axes', titlesize=large_fontSize)
        plt.rc('axes', labelsize=large_fontSize)
        plt.rc('legend', fontsize=small_fontSize)
        plt.rc('xtick', labelsize=small_fontSize)
        plt.rc('ytick', labelsize=small_fontSize) 

        def myplot(score, coeff, labels=None):
            xs = score[:,0]
            ys = score[:,1]
            n = coeff.shape[0]
            scalex = 1.0/(xs.max() - xs.min())
            scaley = 1.0/(ys.max() - ys.min())
            scatt_X = xs * scalex
            scatt_Y = ys * scaley
            scatter = plt.scatter(
                scatt_X, 
                scatt_Y, 
                alpha=0.8, 
                label='Date samples',
            )

            for i in range(n):
                arrow = plt.arrow(
                    0, 0, coeff[i,0], coeff[i,1], color = 'r', alpha = 0.9, 
                    head_width=0.05, head_length=0.05, label='Loadings',
                )
                if labels is None:
                    plt.text(
                        coeff[i,0] * 1.15, coeff[i,1] * 1.15, "Var"+str(i+1), 
                        color='g', ha='center', va='center',
                    )
                else:
                    plt.text(
                        coeff[i,0] * 1.15, coeff[i,1] * 1.15, labels[i], 
                        color='g', ha='center', va='bottom',
                    )

            if(show_labels):
                for x_pos, y_pos, label in zip(scatt_X, scatt_Y, piv.dropna().index.date):
                    ax.annotate(
                        label, # The label for this point
                        xy=(x_pos, y_pos), # Position of the corresponding point
                        xytext=(7, 0),     # Offset text by 7 points to the right
                        textcoords='offset points', # tell it to use offset points
                        ha='left',         # Horizontally aligned to the left
                        va='center',       # Vertical alignment is centered
                        color='black', 
                        alpha=0.8,
                    )
            plt.legend(
                [scatter, arrow], 
                ['Date samples', 'Loadings'],
            )

        samples = x_new.shape[0] * piv.shape[1]
        
        # Add the meta-data information box
        props = dict(boxstyle='round', facecolor='grey', alpha=0.15)      
        ax.text(
            1.1, 
            0.5, 
            # 'Start date:  {}\nEnd date:    {}\n\nOriginal samples:     {}\nSamples used:     {}\nDate samples:               {}'.format(piv.index[0].date(), piv.index[-1].date(), totalSamples, samples, x_new.shape[0]), 
            f"Start date: {piv.index[0].date()}\nEnd date: {piv.index[-1].date()}\n\nOriginal samples: {totalSamples}\nSamples used: {samples}\nDate samples: {x_new.shape[0]}",
            transform=ax.transAxes, 
            fontsize=20, 
            fontweight='bold', 
            verticalalignment='bottom', 
            bbox=props,
        )

        plt.xlim(-1,1)
        plt.ylim(-1,1)
        plt.xlabel("PC{}".format(1))
        plt.ylabel("PC{}".format(2))
        ax.set_title(title, fontweight='bold')
        plt.grid(alpha=0.5)

        #Call the function. Use only the 2 Principal Components.
        myplot(
            x_new[:,0:2], 
            np.transpose(pca.components_[0:2, :]), 
            labels=piv.columns,
        )
        plt.show()

        # Save figure (if specified)
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        fig.savefig(
            f"{save_dir}/{title}.png", bbox_inches="tight",
            # save_dir + '/' + title +'.png', bbox_inches="tight"
        )


# Logger Details
# -------------------------------------------------
LOGFILE_DIR = "./logs"

# LOG_FORMAT = "%(asctime)s [%(levelname)s] File %(module)s: line %(lineno)d: %(message)s"
# LOG_FORMAT = "%(asctime)s [%(levelname)s] File %(pathname)s: line %(lineno)d:\n%(message)s"
# LOG_FORMAT = "%(asctime)s [%(levelname)s] File %(pathname)s: line %(lineno)d: %(message)s"
LOG_FORMAT = "%(asctime)s [%(levelname)s] %(filename)s: line %(lineno)d: %(message)s"

LOG_SEPARATOR = f"\n\n\n{'-'*50}\n\n\n"



# Data Details
# -------------------------------------------------
REQUIRED_DATA_COLUMNS = [
    'COLLECTION_DATE',
    'STATION_ID',
    'ANALYTE_NAME',
    'RESULT',
    'RESULT_UNITS',
]

REQUIRED_CONSTRUCTION_DATA_COLUMNS = [
    'STATION_ID',
    'AQUIFER',
    'WELL_USE',
    'LATITUDE',
    'LONGITUDE',
    'GROUND_ELEVATION',
    'TOTAL_DEPTH',
]

COLLECTION_DATE_FORMAT = "%m/%d/%y"



# Coordinates information
# -------------------------------------------------
DEFAULT_SOURCE_COORDINATES = (436642.70, 3681927.09)




# Requirement Messages
# -------------------------------------------------
PYLENM_DATA_REQUIREMENTS = f"""
PYLENM DATA REQUIREMENTS:
The imported data needs to meet ALL of the following conditions to have a successful import:
    1) Data should be a pandas dataframe.
    2) Data must have these column names: {REQUIRED_DATA_COLUMNS}
"""

PYLENM_CONSTRUCTION_DATA_REQUIREMENTS = f"""
PYLENM CONSTRUCTION DATA REQUIREMENTS:
The imported data needs to meet ALL of the following conditions to have a successful import:
    1) Data should be a pandas dataframe.
    2) Data must have these column names: {REQUIRED_CONSTRUCTION_DATA_COLUMNS}
"""




import numpy as np

from sklearn.model_selection import GridSearchCV
from sklearn.ensemble import RandomForestRegressor
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern, WhiteKernel
from sklearn.linear_model import (
    LinearRegression, 
    Ridge, 
    Lasso, 
    RidgeCV, 
    LassoCV,
)

from pylenm2.utils import constants as c

import logging
from pylenm2 import logger_config

gp_logger = logger_config.setup_logging(
    module_name=__name__,
    # level=logging.INFO,
    level=logging.DEBUG,
    logfile_dir=c.LOGFILE_DIR,
)


def get_Best_GP(
        # self, 
        X, 
        y, 
        smooth=True, 
        seed=42,
    ):
    """Returns the best Gaussian Process model for a given X and y.
    TODO: ASK: Ask if the bound values for `Matern` and `WhiteKernel` are chosen particularly or arbitrarily. @Zexuan
    
    Args:
        X (numpy.array): array of dimension (number of stations, 2) where each element is a pair of UTM coordinates.
        y (numpy.array): array of size (number of stations) where each value corresponds to a concentration value at a station.
        smooth (bool, optional): flag to toggle WhiteKernel on and off. Defaults to True.
        seed (int, optional): random state setting. Defaults to 42.

    Returns:
        GaussianProcessRegressor: best GP model
    """

    gp = GaussianProcessRegressor(normalize_y=True, random_state=seed, optimizer='fmin_l_bfgs_b')
    
    # Kernel models
    if(smooth):
        k1 = Matern(length_scale=400, nu=1.5, length_scale_bounds=(100.0, 5000.0)) + WhiteKernel() #noise_level_bounds=(1e-10, 1e5))
        k2 = Matern(length_scale=800, nu=1.5, length_scale_bounds=(100.0, 5000.0)) + WhiteKernel() #noise_level_bounds=(1e-10, 1e5))
        k3 = Matern(length_scale=1200, nu=1.5, length_scale_bounds=(100.0, 5000.0)) + WhiteKernel() #noise_level_bounds=(1e-10, 1e5))
        k4 = Matern(length_scale=400, nu=1.5, length_scale_bounds=(100.0, 5000.0)) + WhiteKernel() #noise_level_bounds=(1e-10, 1e5))
        k5 = Matern(length_scale=800, nu=1.5, length_scale_bounds=(100.0, 5000.0)) + WhiteKernel() #noise_level_bounds=(1e-10, 1e5))
        k6 = Matern(length_scale=1200, nu=2.5, length_scale_bounds=(100.0, 5000.0)) + WhiteKernel() #noise_level_bounds=(1e-10, 1e5))
        k7 = Matern(length_scale=400, nu=2.5, length_scale_bounds=(100.0, 5000.0)) + WhiteKernel() #noise_level_bounds=(1e-10, 1e5))
        k8 = Matern(length_scale=800, nu=2.5, length_scale_bounds=(100.0, 5000.0)) + WhiteKernel() #noise_level_bounds=(1e-10, 1e5))
        k9 = Matern(length_scale=1200, nu=2.5, length_scale_bounds=(100.0, 5000.0)) + WhiteKernel() #noise_level_bounds=(1e-10, 1e5))
        k10 = Matern(length_scale=400, nu=np.inf, length_scale_bounds=(100.0, 5000.0)) + WhiteKernel() #noise_level_bounds=(1e-10, 1e5))
        k11 = Matern(length_scale=800, nu=np.inf, length_scale_bounds=(100.0, 5000.0)) + WhiteKernel() #noise_level_bounds=(1e-10, 1e5))
        k12 = Matern(length_scale=1200, nu=np.inf, length_scale_bounds=(100.0, 5000.0)) + WhiteKernel() #noise_level_bounds=(1e-10, 1e5))
    else:
        k1 = Matern(length_scale=400, nu=1.5, length_scale_bounds=(100.0, 5000.0))
        k2 = Matern(length_scale=800, nu=1.5, length_scale_bounds=(100.0, 5000.0))
        k3 = Matern(length_scale=1200, nu=1.5, length_scale_bounds=(100.0, 5000.0))
        k4 = Matern(length_scale=400, nu=1.5, length_scale_bounds=(100.0, 5000.0))
        k5 = Matern(length_scale=800, nu=1.5, length_scale_bounds=(100.0, 5000.0))
        k6 = Matern(length_scale=1200, nu=2.5, length_scale_bounds=(100.0, 5000.0))
        k7 = Matern(length_scale=400, nu=2.5, length_scale_bounds=(100.0, 5000.0))
        k8 = Matern(length_scale=800, nu=2.5, length_scale_bounds=(100.0, 5000.0))
        k9 = Matern(length_scale=1200, nu=2.5, length_scale_bounds=(100.0, 5000.0))
        k10 = Matern(length_scale=400, nu=np.inf, length_scale_bounds=(100.0, 5000.0))
        k11 = Matern(length_scale=800, nu=np.inf, length_scale_bounds=(100.0, 5000.0))
        k12 = Matern(length_scale=1200, nu=np.inf, length_scale_bounds=(100.0, 5000.0))
    
    parameters = {'kernel': [k1,k2,k3,k4,k5,k6,k7,k8,k9,k10,k11,k12]}
    
    model = GridSearchCV(gp, parameters)
    model.fit(X, y)
    
    return model


def fit_gp(
        # self, 
        X, 
        y, 
        xx, 
        model=None, 
        smooth=True,
    ):
    """Fits Gaussian Process for X and y and returns both the GP model and the predicted values

    Args:
        X (numpy.array): array of dimension (number of stations, 2) where each element is a pair of UTM coordinates.
        y (numpy.array): array of size (number of stations) where each value corresponds to a concentration value at a station.
        xx (numpy.array): prediction locations
        model (GaussianProcessRegressor, optional): model to fit. Defaults to None.
        smooth (bool, optional): flag to toggle WhiteKernel on and off. Defaults to True.

    Returns:
        GaussianProcessRegressor, numpy.array: GP model, prediction of xx
    """
    
    if(model==None):
        gp = get_Best_GP(X, y, smooth) # selects best kernel params to fit
    else:
        gp = model
    
    gp.fit(X, y)
    y_pred = gp.predict(xx)
    
    return gp, y_pred


def interpolate_topo(
        # self, 
        X, 
        y, 
        xx, 
        ft=['Elevation'], 
        model=None, 
        smooth=True, 
        regression='linear', 
        seed = 42,
    ):
    """Spatially interpolate the water table as a function of topographic metrics using Gaussian Process. Uses regression to generate trendline adds the values to the GP map.

    Args:
        X (numpy.array): training values. Must include "Easting" and "Northing" columns.
        y (numpy.array): array of size (number of stations) where each value corresponds to a concentration value at a station.
        xx (numpy.array): prediction locations
        ft (list, optional): eature names to train on. Defaults to ['Elevation'].
        model (GaussianProcessRegressor, optional): model to fit. Defaults to None.
        smooth (bool, optional): flag to toggle WhiteKernel on and off. Defaults to True.
        regression (str, optional): choice between 'linear' for linear regression, 'rf' for random forest regression, 'ridge' for ridge regression, or 'lasso' for lasso regression.. Defaults to 'linear'.
        seed (int, optional): random state setting. Defaults to 42.

    Returns:
        numpy.array: predicton of locations xx
    """

    # `alpha_values` are only used for RidgeCV and LassoCV models
    alpha_Values = [1e-5, 5e-5, 0.0001, 0.0005, 0.005, 0.05, 0.1, 0.3, 1, 3, 5, 10, 15, 30, 80]
    
    if(regression.lower()=='linear'):
        reg = LinearRegression()
    
    if(regression.lower()=='rf'):
        reg = RandomForestRegressor(n_estimators=200, random_state=seed)
    
    if(regression.lower()=='ridge'):
        # reg = make_pipeline(PolynomialFeatures(3), Ridge())
        reg = RidgeCV(alphas=alpha_Values)
    
    if(regression.lower()=='lasso'):
        # reg = make_pipeline(PolynomialFeatures(3), Lasso())
        reg = LassoCV(alphas=alpha_Values)
    
    if(all(elem in list(xx.columns) for elem in ft)):
        reg.fit(X[ft], y)
        y_est = reg.predict(X[ft])
        residuals = y - y_est
        if(model==None):
            model = get_Best_GP(X[['Easting','Northing']], residuals, smooth=smooth, seed=seed)
        else:
            model = model
        reg_trend = reg.predict(xx[ft])
    else:
        reg.fit(X[['Easting','Northing']], y)
        y_est = reg.predict(X[['Easting','Northing']])
        residuals = y - y_est
        if(model==None):
            model = get_Best_GP(X[['Easting','Northing']], residuals, smooth=smooth, seed=seed)
        else:
            model = model
        reg_trend = reg.predict(xx)
    
    r_map = model.predict(xx[['Easting','Northing']])
    y_map = reg_trend + r_map
    
    return y_map, r_map, residuals, reg_trend


import pandas as pd
import matplotlib.dates as mdates
from sklearn.cluster import KMeans
from matplotlib import pyplot as plt
from matplotlib.dates import date2num

from pylenm2.data import filters
from pylenm2.utils import constants as c

import logging
from pylenm2 import logger_config

cluster_logger = logger_config.setup_logging(
    module_name=__name__,
    # level=logging.INFO,
    level=logging.DEBUG,
    logfile_dir=c.LOGFILE_DIR,
)


# NOT in use!
def __cluster_data_OLD(self, data, n_clusters=4, log_transform=False, filter=False, filter_station_by=['D'], return_clusters=False):
    if(filter):
        res_stations = self.filter_stations(filter_station_by)
        data = data.T
        data = data.loc[data.index.isin(res_stations)]
        data = data.T
    if(log_transform):
        data = np.log10(data)
        data = data.dropna(axis=1)
    temp = data.T
    k_Means = KMeans(n_clusters=n_clusters, random_state=42)
    km = k_Means.fit(temp)
    predict = km.predict(temp)
    temp['predicted'] = km.labels_
    colors = ['red', 'blue', 'orange', 'purple', 'green', 'beige', 'pink', 'black', 'cadetblue', 'lightgreen']
    temp['color'] = temp['predicted'].map(lambda p: colors[p])

    fig, ax = plt.subplots(figsize=(20,10))
    ax = plt.axes()

    color = temp['color']
    for x in range(temp.shape[0]):
        curr = data.iloc[:,x]
        ax.plot(curr, label=curr.name, color=color[x])
        ax.legend()
    
    if(return_clusters):
        color_df = pd.DataFrame(temp['color'])
        color_df['STATION_ID'] = color_df.index
        if(self.get_Construction_Data==None):
            print('You need to set the GPS data first using the setConstructionData function.')
            return None
        else:
            gps_color = pd.merge(self.get_Construction_Data(), color_df, on=['STATION_ID'])
            return gps_color


def cluster_data(
        # self, 
        data_pylenm_dm,
        data, 
        analyte_name=["ANALYTE_NAME"], 
        n_clusters=4, 
        filter=False, 
        col=None, 
        equals=[], 
        year_interval=5, 
        y_label = 'Concentration', 
        return_clusters=False,
    ):
    """Clusters time series concentration data using kmeans algorithm and plots it.

    Args:
        data_pylenm_dm (pylenm2.PylenmDataModule): PylenmDataModule object containing the concentration and construction data.
        data (pd.DataFrame): data to be used in clustering.
        analyte_name (list, optional): analytes to use to cluster. Defaults to ["ANALYTE_NAME"].
        n_clusters (int, optional): number of clusters for kmeans. Defaults to 4.
        filter (bool, optional): flag to indicate filtering. Defaults to False.
        col (str, optional): column to filter. Example: col='STATION_ID'. Defaults to None.
        equals (list, optional): values to filter col by. Examples: equals=['FAI001A', 'FAI001B']. Defaults to [].
        year_interval (int, optional): plot x_label interval in years. Defaults to 5.
        y_label (str, optional): y axis label. Defaults to 'Concentration'.
        return_clusters (bool, optional): flag to return cluster assignemnt. Defaults to False.
    """
    
    data = data.copy()

    # Filter data (if specified)
    if(filter):
        # filter_res = self.filter_by_column(data=self.get_Construction_Data(), col=col, equals=equals)
        filter_res = filters.filter_by_column(
            data=data_pylenm_dm.get_construction_data(), 
            col=col, 
            equals=equals,
        )

        # if('ERROR:' in str(filter_res)):
        #     return filter_res
        if filter_res is None:
            return None
        
        query_stations = list(data.columns)
        filter_stations = list(filter_res.index.unique())
        intersect_stations = list(set(query_stations) & set(filter_stations))
        
        cluster_logger.info(f"{intersect_stations = }")
        cluster_logger.info(f"{len(intersect_stations) = }")
        
        if len(intersect_stations) <= 0:
            # return 'ERROR: No results for this query with the specifed filter parameters.'
            cluster_logger.error("ERROR: No results for this query with the specifed filter parameters.")
            return None

        data = data[intersect_stations]
    
    data.index = date2num(data.index)
    temp = data.T

    # KMeans clustering
    k_Means = KMeans(n_init="auto", n_clusters=n_clusters, random_state=43)
    km = k_Means.fit(temp)
    predict = km.predict(temp)
    
    temp['predicted'] = km.labels_
    colors = [
        'red', 'blue', 'orange', 'purple', 'green', 'pink', 
        'black', 'cadetblue', 'lightgreen','beige',
    ]
    temp['color'] = temp['predicted'].map(lambda p: colors[p])
    
    # Generate plots
    fig, ax = plt.subplots(figsize=(10,10), dpi=100)
    ax.minorticks_off()
    # ax = plt.axes()

    color = temp['color']
    for x in range(temp.shape[0]):
        curr = data.iloc[:,x]
        ax.plot(curr, label=curr.name, color=color[x])
        
    years = mdates.YearLocator(year_interval)  # every year
    months = mdates.MonthLocator()  # every month
    yearsFmt = mdates.DateFormatter('%Y') 
    
    ax = plt.gca()
    ax.xaxis.set_major_locator(years)
    ax.xaxis.set_major_locator(years)
    ax.xaxis.set_major_formatter(yearsFmt)
    ax.autoscale_view()
    
    ax.set_xlabel("Years", fontsize=20)
    plt.xticks(fontsize=20)
    
    ax.set_ylabel(y_label, fontsize=20)
    plt.yticks(fontsize=20)
    
    # ax.set_title("{}: {} clusters".format(analyte_name, n_clusters), fontsize=20)
    ax.set_title(f"{analyte_name}: {n_clusters} clusters", fontsize=20)
    
    # Return clusters (if specified)
    if(return_clusters):
        color_df = pd.DataFrame(temp['color'])
        color_df['STATION_ID'] = color_df.index
        
        if data_pylenm_dm.get_construction_data() is None:
            cluster_logger.error("You need to set the GPS data first using the setConstructionData function.")
            return None
        else:
            gps_color = pd.merge(
                data_pylenm_dm.get_construction_data(), 
                color_df, 
                on=['STATION_ID'],
            )
            return gps_color
    
    # if(return_data):
    #     return color, temp


import setuptools
from setuptools import setup
import io
import os
import re


with open("README.md", "r") as fh:
    long_description = fh.read()


def read(path, encoding='utf-8'):
    path = os.path.join(os.path.dirname(__file__), path)
    with io.open(path, encoding=encoding) as fp:
        return fp.read()


def version(path):
    version_file = read(path)
    version_match = re.search(r"""^__version__ = ['"]([^'"]*)['"]""",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='pylenm',
    version=version('pylenm2/version.py'),
    description='This package aims to provide machine learning (ML) functions for performing comprehensive soil and groundwater data analysis, and for supporting the establishment of effective long-term monitoring.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/ALTEMIS-DOE/pylenm',
    author='Aurelien Meray',
    author_email='aurelien.meray@gmail.com',
    license='MIT',
    packages=setuptools.find_packages(),
    install_requires=[
        'markdown',
        'pandas',
        'openpyxl>=2.6.0',
        'pyproj>=3.0.1',
        # 'richdem',
        'rasterio',
        'numpy',
        'seaborn',
        'matplotlib',
        'statsmodels',
        'scipy',
        'pyproj',
        'datetime',
        'scikit-learn>=0.24.1',
        'supersmoother',
        'ipyleaflet>=0.13.0',
        'ipywidgets>=7.5.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    zip_safe=False,
    project_urls={
        "Bug Tracker": "https://github.com/ALTEMIS-DOE/pylenm/issues",
        "Documentation": "https://pylenm.readthedocs.io/",
        "Source": "https://github.com/ALTEMIS-DOE/pylenm",
    },
)